/*$Id: acre_make_inputs.c,v 1.1 2004/04/30 Joel M. Malard Exp d3j540 $ */

#ifdef PETSC_RCS_HEADER
static char vcid[] =
"$Id: acre_make_inputs.c,v 1.1 2004/04/30 Joel M. Malard Exp d3j540 $";
#endif

static char help[] = 
"Initializes a dense matrix A, a vector b and a sparse matrix L.\n\
 Input arguments are:\n\
       -acre_ascii : all output files are either binary or ascii.\n\n\
       -acre_b <int> : size of blocks along the diagonal of L. \n\n\
       -acre_e [band|block] : type of test case to construct.\n\n\
       -acre_i <nits> : number of non-linear iteration steps.\n\n\
       -acre_fA <filename> : the name of the file containing the dense\n\
       exogenous matrix A. The matrix A is of type MATMPIDENSE (PETSc).\n\n\
       -acre_fb <filename> : the name of the file containing the response\n\
       vector.\n\n\
       -acre_fsL <filename> : the name of the file containing the sparsity\n\
       structure of L. The structural file is of type MATMPIAIJ (PETSc)\n\n\
       -acre_flb <filename> : the name of an optional file containing\n\
       lower bounds on the parameters.\n\n\
       -acre_fub <filename> : the name of an optional file containing\n\
       upper bounds on the parameters.\n\n\
       -acre_fx <filename> : the name of an optional file containing an\n\
       initial guess of the parameters.\n\n\
       -acre_help : print this message and exit.\n\n\
       -acre_m <eqns> : number of rows of A (equations).\n\n\
       -acre_n <parm> : number of columns of A (parameters/effects).\n\n\
       -acre_q <comp> : number of variance components to estimate.\n\
       Only used for the test cases of type band.\n\n\
       -acre_std <std> : standard deviation of the noise added to the \n\
       response vector.\n\n";

/* include files */
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "tao.h"
#include "acre.h"
#include "copyright.h"

#define APP_BAND	0
#define APP_BLOCK	1

#undef __FUNC__
#define __FUNC__ "main"
int main(int argc,char **args)
{
	acre_cntxt cntxt;

	Vec     b;		/* response vector */
	Vec     x;           /* vector of parameters */
	Vec     r;           /* residual vector */
	Vec     u;           /* normalized residual vector, Ax+Cu=b */
	Vec     cVec;        /* Variance Components */
	Vec	lBounds;	/* lower bounds on variance components */
	Vec	uBounds;	/* upper bounds on variance components */
	Mat     A;           /* dense data matrix */
	Mat     sL;          /* structure of the matrix L */
	PetscScalar	noise=0.0; /* how much noise to ad to the response */

	int     n = 1;       /* number of columns of A (global & local) */
	int     m = 10;      /* number of rows of A (global) */
	int     nq = 1;      /* number of variance components */
	int     nb = 1;      /* block size for the block diagonal test */
	int	rn;		/* number of local columns of A */
	int	rs;		/* index of first local columns of A */
	int	re;		/* upper bound for local columns indices */
	char    etype[8];    /* example type */
	int     here;        /* processor global rank */
	int     cank, rank;        /* processor group rank */
	int     size;        /* size of MPI_COMM_WORLD */
	int     xc=1;

	int     example;

	/** 
	** Other declarations
	**/
	PetscRandom rState;  /* random number object */
	PetscScalar  one = 1.0;
	PetscScalar  ten = 10.0, onehalf=0.5;
	PetscScalar  mean, stdev;
	char	infile[3000];
	
	int idone;
	int nPrecomp;

	int     *rows;
	PetscScalar  *vals;
	int     *iDiags;
	int     *idc, *idr;
	int     h, i, i0, i1, j, ierr;
	int     start, end, icnt;
	PetscTruth     format, flg;
	int     ustart, uend, unrows;
	MPI_Comm rcomm;

	/***
	 *** Initialize MPI.
	 ***/
	ierr = MPI_Init(&argc, &args);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_rank ( MPI_COMM_WORLD, &here ); 
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_size ( MPI_COMM_WORLD, &size ); 
		assert(ierr==MPI_SUCCESS);

	/***
	 ***  Start up PETSC and find out the command line arguments
	 ***/
	TaoInitialize(&argc,&args,(char *)0,help);

	/**** Check whether or not to print the copyright notice. 
	 ****/
	ierr = PetscOptionsHasName(PETSC_NULL, "-acre_silent",
		&flg);						CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
	  PetscPrintf(PETSC_COMM_SELF,
	  "\n%s%s\n%s\n%s%s%s%s\n%s%s%s%s\n%s%s%s\n",
	  "Copyright 1997 University of Edinburgh\n",
	  "Copyright 2002, 2004 Battelle Memorial Institute\n",
	  "This file is part of the ACRE version 1.1.\n",
	  "ACRE is free software; you can redistribute it and/or\n",
	  "modify it under the terms of the GNU General Public License\n",
	  "as published by the Free Software Foundation; either version 2\n",
	  "of the License, or (at your option) any later version.\n",
	  "ACRE is distributed in the hope that it will be useful,\n",
	  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n",
	  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n",
	  "GNU General Public License for more details.\n",
	  "You should have received a copy of the GNU General Public License\n",
	  "along with ACRE; if not, write to the Free Software Foundation,\n",
	  "Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA\n");
	}

	ierr = PetscOptionsHasName(PETSC_NULL,"-acre_help",
		&flg);						CHKERRQ(ierr);
	if ( flg==PETSC_TRUE )
	{
		PetscPrintf(MPI_COMM_WORLD,"ACRE USAGE\n%s\n",
			help);
		ierr = TaoFinalize();				CHKERRQ(ierr);
		exit(0);
	}

	/****
	 **** Argument -acre_m is the number of rows of the
	 **** exogenous matrix A.
	 ****/
	ierr = PetscOptionsGetInt(TAO_NULL,"-acre_m",&m,&flg);	CHKERRQ(ierr);
	if(here==0) PetscPrintf(PETSC_COMM_SELF,
		"-acre_m= %d\n",m);

	/****
	 **** Argument -acre_n is the number of columns of
	 **** the exogenous matrix A. It is forced to be
	 **** at least as large as the value of -acre_m.
	 ****/
	ierr = PetscOptionsGetInt(TAO_NULL,"-acre_n",&n,&flg);	CHKERRQ(ierr);
	if(n>m) n=m;
	if(here==0) PetscPrintf(PETSC_COMM_SELF,
		"-acre_n= %d\n",n);

	cntxt.ofcn = ACRE_REML;

	/****
	 **** This driver generates two types of covariance matrices, 
	 **** namely: banded and block diagonal. Use the argument
	 **** -acre_e to specify which type.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL,"-acre_e",
		etype,8,&flg);					CHKERRQ(ierr);
	if (flg==PETSC_FALSE || !strcasecmp(etype,"band"))
	{
		example = APP_BAND;
		ierr = PetscOptionsGetInt(TAO_NULL,"-acre_q",
			&nq,&flg);				CHKERRQ(ierr);
		nb = m;
		ierr = PetscOptionsGetInt(TAO_NULL,"-acre_b",
			&nb,&flg);				CHKERRQ(ierr);
	} else
	{
		example = APP_BLOCK;
		ierr = PetscOptionsGetInt(TAO_NULL,"-acre_b",
			&nb,&flg);				CHKERRQ(ierr);
		nq = (nb*nb+nb)/2;
	}
	if(here==0) PetscPrintf(PETSC_COMM_SELF,
		"-acre_b= %d\n",nb);
	if(here==0) PetscPrintf(PETSC_COMM_SELF,
		"-acre_q= %d\n",nq); 

	ierr = MPI_Barrier ( MPI_COMM_WORLD ); 
		assert(ierr==MPI_SUCCESS);
	ierr = MPI_Comm_rank ( PETSC_COMM_WORLD, &cank ); 
		assert(ierr==MPI_SUCCESS);


	ierr = PetscRandomCreate(PETSC_COMM_WORLD,
		RANDOM_DEFAULT, &rState);			CHKERRQ(ierr);

	/* check the group communicator and compute the process
	 * group rank.
	 */
	if ( size%xc!=0 )
	{
		if(here==0)
			PetscPrintf(PETSC_COMM_SELF,
				"%s %d %s %s %d\n",
				"The number of process groups", xc,
				"is not an integral divisor of",
				"the number of processors",
				size);
		ierr = TaoFinalize();				CHKERRQ(ierr);
		exit(0);
	}
	if(here==0)
		PetscPrintf(PETSC_COMM_SELF,
			"parallel theta_k tasks = %d\n",
			xc);
	ierr = MPI_Comm_split(MPI_COMM_WORLD, here/xc,
		here%xc, &rcomm);
		assert(ierr==MPI_SUCCESS);
	ierr = MPI_Comm_rank ( rcomm, &rank ); 
		assert(ierr==MPI_SUCCESS);

	rs = rank*(n/xc);
	rs += ( rank < n%xc ) ? rank : (n%xc);

	re = (rank+1)*(n/xc);
	re += ( (rank+1) < n%xc ) ? (rank+1) : (n%xc);
	if ( rank+1==xc ) re = n;
	rn = re - rs;
	
	/** 
	** Create the solution vector, and gradient 
	** used by TAO
	**/
	ierr = VecCreateSeq ( PETSC_COMM_SELF, nq, &cVec);	CHKERRQ(ierr);

	/** Note that columns indices start at 1 to expose
	** any sparsity in sL. This test covariance matrix 
	** is banded so really sL is dense.
	**/

	nPrecomp = ( example==APP_BAND ) ? nq : nb;

	ierr = MatCreateMPIAIJ ( PETSC_COMM_WORLD,
		PETSC_DECIDE, PETSC_DECIDE, m, nq,
		nPrecomp, TAO_NULL, nPrecomp,
		TAO_NULL, &sL);					CHKERRQ(ierr);

	ierr = MatGetOwnershipRange(sL,&start,&end);		CHKERRQ(ierr);

	ierr = PetscMalloc(nq*sizeof(int),&iDiags);		CHKERRQ(ierr);
	for ( i=0;i<nq; i++ ) iDiags[i] = 0;

	if ( example==APP_BAND )
	{

	/* Define a lower triangular banded structure 
	* with a different entry along each diagonal.
	*/
		if(here==0)
			PetscPrintf(PETSC_COMM_SELF,"example: band\n");

		iDiags[0] = 1;

		ierr = PetscMalloc(nq*sizeof(PetscScalar),
			&vals);					CHKERRQ(ierr);
		ierr = PetscMalloc(nq*sizeof(int),
			&idc);					CHKERRQ(ierr);
		for ( i=start; i<end; i++ )
		{
			icnt = 0;
			for( j=0; j<nq && j<=i%nb; j++ )
			{
				vals[icnt] = i-j+1;
				idc[icnt++] = j;
			}
			ierr = MatSetValues( sL, 1, &i, icnt, idc,
				vals, INSERT_VALUES);		CHKERRQ(ierr);
		}
		ierr = MatAssemblyBegin(sL,MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);
		ierr = MatAssemblyEnd(sL,MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);

		PetscFree(vals);
		PetscFree(idc);

	} else
	{

	/* define a plain lower block diagonal structure where the
	* the same nb by nb block is repeated and each entry of
	* this block is free.
	*/

		if(here==0)
			PetscPrintf(PETSC_COMM_SELF,
				"example: block\n");
		ierr = PetscMalloc(nq*sizeof(PetscScalar),
			&vals);					CHKERRQ(ierr);
		ierr = PetscMalloc(sizeof(int), &idr);		CHKERRQ(ierr);
		ierr = PetscMalloc(nq*sizeof(int), &idc);	CHKERRQ(ierr);
		i0 = (start/nb)*nb;
		i1 = start-i0;
		h = (i1*(i1+1))/2;
		for ( i=start; i<end; i++ )
		{
			if(i%nb==0) h=0;
			i0 = (i/nb)*nb;
			icnt = i-i0+1;
			for( j=i0; j<=i; j++ )
			{
				if ( i==j ) iDiags[h] = 1;
				idc[j-i0] = h++;
				vals[j-i0] = j+1;
			}
			idr[0] = i;
			ierr = MatSetValues( sL, 1, idr, icnt, 
				idc, vals, INSERT_VALUES);	CHKERRQ(ierr);
	
		}
		ierr = MatAssemblyBegin(sL,MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);
		ierr = MatAssemblyEnd(sL,MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);

		PetscFree(vals);
		PetscFree(idc);
		PetscFree(idr);

	}

	/**
	** Create the dense data matrix A and response vector b
	**/
	ierr = MatCreateMPIDense(PETSC_COMM_WORLD,
		PETSC_DECIDE, PETSC_DECIDE, m, rn,
		TAO_NULL, &A);					CHKERRQ(ierr);

	ierr = VecCreateMPI( PETSC_COMM_WORLD, PETSC_DECIDE,
		m, &b);						CHKERRQ(ierr);
	/**
	** Create the vector of parameters x and the error vector u.
	**/
	ierr = VecCreateMPI( PETSC_COMM_WORLD, PETSC_DECIDE,
		rn, &x);					CHKERRQ(ierr);
	ierr = VecDuplicate( b, &u);				CHKERRQ(ierr);

	/**
	** Compute a random matrix A and a response vector b. 
	**/
	ierr = MatGetOwnershipRange( A, &start, &end);		CHKERRQ(ierr);

	ierr = PetscMalloc(n*sizeof(PetscScalar),&vals);	CHKERRQ(ierr);
	ierr = PetscMalloc(sizeof(int),&idr);			CHKERRQ(ierr);
	ierr = PetscMalloc(n*sizeof(int),&idc);			CHKERRQ(ierr);
	for ( j=0; j<n; j++ ) idc[j] = j;
	for ( i=0; i<m; i++ )
	{
		if ( here==0 )
		{
			for ( j=0; j<n; j++ )
			{
				/* vals[j] = (i==j) ? 1 : 0; */
	    			ierr = PetscRandomGetValue( rState,
					&vals[j]);
			}
		}
		ierr = MPI_Bcast(vals, n, MPI_DOUBLE, 0,
			MPI_COMM_WORLD );
			assert(ierr==MPI_SUCCESS);
		/* don't forget that columns don't have global indices
		 * when xc>1.
		 */
		if ( i>=start && i<end )
		{
			idr[0]=i;
			ierr = MatSetValues( A, 1, idr,
				rn, idc, vals+rs,
				INSERT_VALUES);			CHKERRQ(ierr);
		}
		if((i%(m/size+1))==0)
		{
			ierr = MatAssemblyBegin( A,
				MAT_FLUSH_ASSEMBLY);		CHKERRQ(ierr);
			ierr = MatAssemblyEnd( A,
				MAT_FLUSH_ASSEMBLY);		CHKERRQ(ierr);
		}
	}
	ierr = MatAssemblyBegin( A, MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);
	ierr = MatAssemblyEnd( A, MAT_FINAL_ASSEMBLY);		CHKERRQ(ierr);
	PetscFree(vals);
	PetscFree(idc);
	PetscFree(idr);

	/***
	*** Bundle all the above into the ACRE context. 
	***/
	acre_start( A, b, sL, cVec, xc, &cntxt);

	/* 
	* Initialize the response vector
	*/
	ierr = VecSet(&one, x);					CHKERRQ(ierr);
	ierr = MatMult( A, x, b);				CHKERRQ(ierr);
	ierr = acre_vector_combine(&cntxt,b);

	/***
	*** Set cVec and define a matrix L with the appropriate 
	*** sparsity structure. The subroutine acre_set_cholesky
	*** reinitializes internal data structures.
	***/
	ierr = PetscMalloc(nq*sizeof(PetscScalar),&vals);	CHKERRQ(ierr);
	ierr = PetscMalloc(nq*sizeof(int),&rows);		CHKERRQ(ierr);
	for ( i=0; i<nq; i++)
		vals[i] = (iDiags[i]!=0) ? 1.0+(double)i/nq: 0.0;
	for ( i = 0; i<nq; i++ ) rows[i] = i;
	ierr = VecSetValues(cVec,nq,rows,vals,INSERT_VALUES);	CHKERRQ(ierr);
	ierr = VecAssemblyBegin( cVec);				CHKERRQ(ierr);
	ierr = VecAssemblyEnd( cVec);				CHKERRQ(ierr);
	PetscFree ( rows);
	PetscFree ( vals);

	ierr = acre_set_cholesky(&cntxt, cVec);			CHKERRQ(ierr);

	/***
	*** Add noise to the response vector b
	***/
	ierr = PetscOptionsGetScalar(PETSC_NULL, "-acre_std",
		&noise, &flg);					CHKERRQ(ierr);
	PetscPrintf(PETSC_COMM_SELF,"noise=%lg\n",noise);
	if ( noise > 0.0 )
	{
		/**
		 ** Compute a random vector with normally distributed
		 ** components of mean 0 and variance noise*noise.
		 **/

		ierr = PetscMalloc(m*sizeof(PetscScalar),
			&vals);					CHKERRQ(ierr);
		if ( here==0 )
		{
			for ( j=0; j<m; j++ )
			{
				ierr = PetscRandomGetValue(
					rState, &vals[j]);	CHKERRQ(ierr);
			}
		}
		ierr = MPI_Bcast(vals, m, MPI_DOUBLE, 0,
			MPI_COMM_WORLD );
			assert(ierr==MPI_SUCCESS);
		ierr = VecGetOwnershipRange(u, &ustart, &uend);	CHKERRQ(ierr);
		unrows = uend-ustart;
		ierr = PetscMalloc(unrows*sizeof(int),&idr);	CHKERRQ(ierr);
		for ( i=ustart; i<uend; i++ ) idr[i-ustart] = i;
		ierr = VecSetValues(u, unrows, idr, &vals[ustart],
			INSERT_VALUES);				CHKERRQ(ierr);
		PetscFree(vals);
		PetscFree(idr);
		ierr = VecAssemblyBegin(u);			CHKERRQ(ierr);
		ierr = VecAssemblyEnd(u);			CHKERRQ(ierr);
		ierr = VecSum( u, &mean);			CHKERRQ(ierr);
		mean = -mean/m;
		ierr = VecDot( u, u, &stdev);			CHKERRQ(ierr);
		stdev /= m-1.0;
		stdev = 1.0/sqrt((double)stdev);
		ierr = VecDuplicate( u, &r);			CHKERRQ(ierr);
		ierr = VecSet( &mean, r);			CHKERRQ(ierr);
		ierr = VecAXPBY(&stdev,&stdev,r,u);		CHKERRQ(ierr);
		ierr = VecScale(&noise,u);			CHKERRQ(ierr);
		ierr = VecDestroy( r);				CHKERRQ(ierr);

		ierr = MatMultAdd( cntxt.L, u, b, b);		CHKERRQ(ierr);
	}

	ierr = VecDuplicate(cVec, &lBounds );			CHKERRQ(ierr);
	idone = 0;
	for ( i=0; i<nq; i++ )
	{
		if ( iDiags[i]==0 )
		{
			VecSetValue(lBounds,i, 0.0,
				INSERT_VALUES);
		}
		else if ( idone==0 )
		{
			VecSetValue(lBounds,i, 1.00 ,
				INSERT_VALUES);
			idone = 1;
		}
		else
		{
			VecSetValue(lBounds,i, 0.01 /*ACRE_SMALL*/,
				INSERT_VALUES);
		}
	}

	ierr = VecDuplicate(cVec, &uBounds );			CHKERRQ(ierr);
	for ( i=0; i<nq; i++ )
	{
		if ( iDiags[i]==0 )
		{
			VecSetValue(uBounds,i, ten,
				INSERT_VALUES);
		}
		else if ( idone==1 )
		{
			VecSetValue(uBounds,i, 1.00 ,
				INSERT_VALUES);
			idone = 0;
		}
		else
		{
			VecSetValue(uBounds,i, ten,
				INSERT_VALUES);
		}
	}

	ierr = VecCopy( lBounds, cVec);				CHKERRQ(ierr);
	ierr = VecAXPY( &one, uBounds, cVec);			CHKERRQ(ierr);
	ierr = VecScale( &onehalf, cVec);			CHKERRQ(ierr);

	ierr = MPI_Barrier(MPI_COMM_WORLD);
		assert(ierr==MPI_SUCCESS);

	ierr = PetscOptionsHasName(PETSC_NULL,"-acre_ascii",
		&format);					CHKERRQ(ierr);

	/**** Argument -acre_fA gives the name of the binary file
	 **** that contains the exogenous matrix A. Each process
	 **** column read the entire matrix and then extracts the
	 **** appropriate columns.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fA",
		infile, 3000,&flg);				CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(here==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_fA is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		exit(0);
	}

	ierr = acre_print_dense_matrix_serial( A, infile,
		format);					CHKERRQ(ierr); 

	/**** Argument -acre_fb gives the name of the binary file
	 **** that contains the response vector. This vector
	 **** is distributed along each processor column and
	 **** replicated along processor rows.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fb", infile,
		3000,&flg);					CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(here==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_fb is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		exit(0);
	}

	ierr = acre_print_vector_serial( b, infile, format);	CHKERRQ(ierr);

	/**** Argument -acre_fsL gives the name of the binary file
	 **** that contains the sparsity structure of the Cholesky
	 **** factor L of the covariance matrix C=LL'. This vector
	 **** is distributed like the response vector b.
	 **** 
	 **** Note: the structural matrix sL has one row for every
	 **** row of A and L and one column for every (co)variance
	 **** component that is sought, see the above reference 
	 **** paper for details. The following example might help.
	 **** Suppose the exogenous matrix A is 5x2, making the
	 **** matrices C and L to be 5x5, with say 3 components t_1,
	 **** t_2 and t_3 to be estimated. The matrix L and its
	 **** structural matrix sL might look like this with 0
	 **** represented by ".":
	 ****
	 ****        /                     \         /           \                 
	 ****        | t_1  .   .   .   .  |         |  1  .  .  |                 
	 ****	L =  |                     |   sL =  |           |                 
	 ****        | t_3 t_2  .   .   .  |         |  .  2  1  |                 
	 ****        |                     |         |           |                 
	 ****        | t_3  .  t_1  .   .  |         |  3  .  1  |                 
	 ****        |                     |         |           |                 
	 ****        | t_3  .   .  t_2  .  |         |  .  4  1  |                 
	 ****        |                     |         |           |                 
	 ****        | t_3  .   .   .  t_1 |         |  5  .  1  |                 
	 ****        \                     /         \           /                 
	 ****
	 **** This implementation assumes that no t_i appears more
	 **** than once on any one row of L. That limitation  should
	 **** not be hard to overcome.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL,"-acre_fsL",infile,
		3000,&flg);					CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(here==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_fsL is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		exit(0);
	}

	ierr = acre_print_aij_matrix_serial( sL, infile,
		format);					CHKERRQ(ierr);

	/**** Argument -acre_fx gives the name of the binary file
	 **** that contains an initial guess on the unknown
	 **** components. If this option is omitted, the algorithm
	 **** starts with all ones, an all zero vector would
	 **** lead to a singular matrix L.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fx", infile,
		3000,&flg);					CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(here==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_fx is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		exit(0);
	}
	ierr = acre_print_vector_serial( cVec, infile, format);	CHKERRQ(ierr);

	/**** Argument -acre_flb gives the name of the binary file
	 **** that contains the lower bounds on the unknown
	 **** components. If this option is omitted, the lower
	 **** bound on all components is zero.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_flb", infile,
		3000,&flg);					CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(here==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_flb is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		exit(0);
	}
	ierr = acre_print_vector_serial( lBounds, infile,
		format);					CHKERRQ(ierr);

	/**** Argument -acre_fub gives the name of the binary file
	 **** that contains the lower bounds on the unknown
	 **** components. If this option is omitted, the lower
	 **** bound on all components is zero.
	 ****/
	ierr = PetscOptionsGetString(TAO_NULL, "-acre_fub",
		infile, 3000,&flg);				CHKERRQ(ierr);
	if ( flg!=PETSC_TRUE )
	{
		if(here==0) PetscPrintf(PETSC_COMM_SELF,
			"Argument -acre_fub is missing\n");
		ierr = TaoFinalize();				CHKERRQ(ierr);
		exit(0);
	}
	ierr = acre_print_vector_serial( uBounds, infile,
		format);					CHKERRQ(ierr);

	ierr = MPI_Barrier(MPI_COMM_WORLD);
		assert(ierr==MPI_SUCCESS);

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	Free any remaining Petsc Object
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	acre_end(&cntxt);

	ierr = VecDestroy(u);					CHKERRQ(ierr);
	ierr = VecDestroy(x);					CHKERRQ(ierr);
	ierr = VecDestroy(b);					CHKERRQ(ierr);
	ierr = MatDestroy(A);					CHKERRQ(ierr);
	
	ierr = MatDestroy(sL);					CHKERRQ(ierr);
	ierr = VecDestroy(cVec);				CHKERRQ(ierr);
	ierr = VecDestroy(lBounds);				CHKERRQ(ierr);
	ierr = VecDestroy(uBounds);				CHKERRQ(ierr);
	ierr = PetscRandomDestroy(rState);			CHKERRQ(ierr);

	ierr = TaoFinalize();					CHKERRQ(ierr);
	ierr = MPI_Finalize();
		assert(ierr==MPI_SUCCESS);
	
	return 0;

}

